package comp2402a2;

import java.util.Comparator;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

public class FastMinDeque<T> implements MinDeque<T> {

  protected Comparator<? super T> comp;

  protected FastMinStack<T> front;
  protected FastMinStack<T> back;

  public FastMinDeque() {
    this(new DefaultComparator<T>());
  }

  public FastMinDeque(Comparator<? super T> comp) {
    this.comp = comp;
    front = new FastMinStack<>();
    back = new FastMinStack<>();
  }

  protected List<T> rebalance() {
      // put all the contents into a list, in the correct order
      List<T> tmp = new ArrayList<>();
      while (front.size() > 0) {
        tmp.add(front.pop());
      }
      for (int i = 0; i < back.size(); i++) {
        tmp.add(null);
      }
      int m = back.size();
      for (int i = 0; i < m; i++) {
        tmp.set(tmp.size()-i-1, back.pop());
      }
      // spread the contents evenly among front and back
      int n = tmp.size();
      for (int i = n/2; i < n; i++) {
        back.push(tmp.get(i));
      }
      for (int i = n/2-1; i >= 0; i--) {
        front.push(tmp.get(i));
      }
      return tmp;  // return the list (useful for making an iterator)
  }

  public void addFirst(T x) {
    front.push(x);
  }

  public void addLast(T x) {
    back.push(x);
  }

  public T removeFirst() {
    if (front.size() == 0) {
      if (back.size() == 1) {
        return back.pop();
      }
      rebalance();
    }
    return front.pop();
  }

  public T removeLast() {
    if (back.size() == 0) {
      rebalance();
    }
    return back.pop();
  }

  public T min() {
    T m = null;
    if (front.size() > 0) {
      if (back.size() > 0 && comp.compare(back.min(), front.min()) < 0) {
        m = back.min();
      } else {
        m = front.min();
      }
    } else {
      m = back.min();
    }
    return m;
  }

  public int size() {
    return front.size() + back.size();
  }

  public Iterator<T> iterator() {
    return new FMDIterator<>(this);
  }

  protected class FMDIterator<T> implements Iterator<T> {
    Iterator<T> fi, bi;

    public FMDIterator(FastMinDeque<T> fmd) {
      fi = fmd.front.all.descendingIterator();
      bi = fmd.back.all.iterator();
    }
    public T next() {
      if (fi.hasNext()) {
        return fi.next();
      }
      return bi.next();
    }

    public boolean hasNext() {
      return fi.hasNext() || bi.hasNext();
    }
  }

  public Comparator<? super T> comparator() {
    return comp;
  }
}
